import sys
import socket
import signal
import os
from .proxy import Proxy, signal_handler
from knack import CLI
from knack.commands import CLICommandsLoader, CommandGroup
from knack.arguments import ArgumentsContext, CLIArgumentType


class CommandLoader(CLICommandsLoader):
    def load_command_table(self, args):
        with CommandGroup(self, "proxy", "aproxy.proxy#{}") as g:
            g.command("single", "start_single_proxy")
            g.command("start", "start_from_config")
        return super(CommandLoader, self).load_command_table(args)

    def load_arguments(self, command):
        with ArgumentsContext(self, "proxy single") as ac:
            ac.argument("local_port", type=int)
            ac.argument("remote_port", type=int)
        return super(CommandLoader, self).load_arguments(command)


name = "aproxy"
cli = CLI(
    cli_name=name,
    config_dir=os.path.expanduser(os.path.join("~", ".{}".format(name))),
    config_env_var_prefix=name,
    commands_loader_cls=CommandLoader,
)
signal.signal(signal.SIGINT, signal_handler)

exit_code = cli.invoke(sys.argv[1:])
sys.exit(exit_code)



from colorama import Fore
from enum import Enum
import hexdump
from .proxy_config import ProxyItem


class Direction(Enum):
    LOCAL = 0
    REMOTE = 1


def get_direction_label(
    direction: Direction, textLabel: str = None, config: ProxyItem = None
):
    if direction == Direction.LOCAL:
        label = "==>"
    else:
        label = "<=="
    label = f"[{textLabel} {Fore.GREEN}{label}{Fore.RESET}]"
    if config and config.provider:
        label = (
            label
            + f" {Fore.YELLOW}({Fore.RESET}{config.provider}{Fore.YELLOW}){Fore.RESET}"
        )
    return label


def print_info(buffer: str, direction: Direction, config: ProxyItem):
    if config.verbosity == 0:
        return
    if not len(buffer):
        return
    label = get_direction_label(direction, config)
    color = Fore.CYAN if direction is Direction.LOCAL else Fore.YELLOW
    print(
        get_direction_label(direction, config.name, config)
        + color
        + f" Received {str(len(buffer))} bytes of data from {direction.name}"
    )

    if config.verbosity > 2:
        hexdump.hexdump(buffer)

    print(Fore.RESET)



import json
import importlib
from aproxy.providers.provider_config import ProviderConfigItem, ProviderConfig
from colorama import Fore


class ProxyConfig:
    def __init__(self, providers: [], proxies: []):
        self.proxies = proxies
        self.providers = providers


class ProxyItem:
    def __init__(
        self,
        local_host: str,
        local_port: int,
        remote_host: str,
        remote_port: int,
        name: str,
        verbosity: int,
        provider: str = None,
    ):
        self.local_host = local_host
        self.local_port = local_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.receive_first = False
        self.name = name
        self.verbosity = verbosity
        self.provider = provider


def dict_to_config(json_config: dict):
    proxies = []
    for item in json_config["proxies"]:
        local_host = item["localHost"] if "localHost" in item else "0.0.0.0"
        local_port = int(item["localPort"]) if "localPort" in item else 0
        remote_host = item["remoteHost"] if "remoteHost" in item else None
        remote_port = int(item["remotePort"]) if "remotePort" in item else 0
        verbosity = int(item["verbosity"]) if "verbosity" in item else 0
        name = item["name"] if "name" in item else "<noname>"
        provider = item["provider"] if "provider" in item else None
        proxy = ProxyItem(
            local_host, local_port, remote_host, remote_port, name, verbosity, provider
        )
        proxies.append(proxy)
    providers = __load_provider_config(json_config["providerConfig"])
    config = ProxyConfig(providers, proxies)
    return config


def __load_provider_config(cfg: dict):
    providers = {}
    for provider in cfg:
        name = provider["name"]
        provider_name = provider["provider"]["name"]
        full_name = "aproxy.providers." + provider_name
        print(
            f"[*] loading provider configuration for {provider_name} ({full_name}) as {name}"
        )
        provider_module = importlib.import_module(full_name)
        provider = provider_module.load_config(provider["provider"])
        provider.connect()
        providers[name] = provider

    return providers


def load_config(configFile: str):
    with open(configFile, "r") as f:
        cfg = json.load(f)
        return dict_to_config(cfg)



import socket, threading, sys, json
from enum import Enum
import hexdump
from colorama import Fore
import select
from .util import Direction, print_info, get_direction_label
from .proxy_config import load_config, ProxyConfig, ProxyItem

running_proxies = {}
stop_proxies = False
config: ProxyConfig = None


def signal_handler(sig, frame):
    global running_proxies
    global stop_proxies

    if running_proxies:
        for proxy in running_proxies.values():
            proxy.stop()
        running_proxies.clear()
    stop_proxies = True
    sys.exit(0)


def start_from_config(configFile: str):
    global config
    config = load_config(configFile)

    print("[*] loaded config for {} proxies".format(len(config.proxies)))
    for proxy in config.proxies:
        thread = threading.Thread(target=start_proxy, args=(proxy,))
        thread.start()


def start_single_proxy(
    local_port: int,
    remote_port: int,
    remote_host: str,
    local_host: str = None,
    verbosity: int = 0,
):
    cfg = ProxyItem(
        local_host, local_port, remote_host, remote_port, "SingleHost", verbosity
    )
    start_proxy(cfg)


def start_proxy(proxy_config: ProxyItem):
    global running_proxies
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    remote_socket: socket.socket = None
    try:
        server.bind((proxy_config.local_host, proxy_config.local_port))
        print(f"[*] Listening on {proxy_config.local_host}:{proxy_config.local_port}")

    except Exception as e:
        print(
            f"[!!] Failed to listen on {proxy_config.local_host}:{proxy_config.local_port}: {str(e)}"
        )
        print(e)
        sys.exit(0)

    server.listen()
    global stop_proxies
    while not stop_proxies:
        client_socket, addr = server.accept()
        print(f"[==>] Received incoming connection from {addr[0]}:{str(addr[1])}")
        proxy = Proxy(client_socket, proxy_config, remote_socket)
        proxy.start()
        running_proxies[proxy.name] = proxy


class Proxy:
    def __init__(
        self,
        client_socket: socket.socket,
        config: ProxyItem,
        remote_host: socket.socket = None,
    ):
        super().__init__()

        self.name = "{}->{}:{}".format(
            client_socket.getsockname(), config.remote_host, config.remote_port
        )
        self.__local = client_socket
        self.__config = config
        self.__stop = False
        if remote_host:
            self.__remote = remote_host
        else:
            self.__remote_connect()

    def start(self):
        self.__thread = threading.Thread(target=self.__proxy_loop)
        self.__thread.start()

    def stop(self):
        if self.__stop:
            return

        self.__stop = True
        if self.__local:
            self.__local.close()
            self.__local = None

        if self.__remote:
            self.__remote.close()
            self.__remote = None

        print(Fore.MAGENTA + "Disconnected " + self.name + Fore.RESET)

    def __remote_connect(self):
        global config
        if self.__config.provider:
            provider = config.providers[self.__config.provider]
            self.__remote = provider.client_connect(
                self.__config.remote_host, self.__config.remote_port, self.__local
            )
        else:
            remote_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            remote_socket.connect(
                (self.__config.remote_host, self.__config.remote_port)
            )
            self.__remote = remote_socket

    def __receive_first(self, receive_first):
        if receive_first:
            remote_buffer = self.__receive_from(self.__remote)
            if self.__dump_data:
                hexdump(remote_buffer)
            remote_buffer = self.__response_handler(remote_buffer)
            if len(remote_buffer):
                print(
                    "[<==] Receiving {} bytes for localhost".format(
                        str(len(remote_buffer))
                    )
                )
                self.__local.send(remote_buffer)

    def __proxy_loop(self):
        self.__receive_first(self.__config.receive_first)
        try:
            while True:
                if self.__stop:
                    break
                self.__handle_traffic(Direction.LOCAL)
                self.__handle_traffic(Direction.REMOTE)
        except Exception:
            import traceback

            print(traceback.format_exc())

    def __handle_traffic(self, direction: Direction):
        if direction == Direction.LOCAL:
            receiver, sender = self.__local, self.__remote
        else:
            receiver, sender = self.__remote, self.__local

        buffer = self.__receive_from(receiver)

        print_info(buffer, direction, self.__config)
        handler = (
            self.__request_handler
            if direction == Direction.LOCAL
            else self.__response_handler
        )

        buffer = handler(buffer)
        if not sender:
            self.__stop = True
            return

        try:
            sender.send(buffer)
        except OSError as e:
            # socket is dead
            self.__stop = True
            return

        if len(buffer) and self.__config.verbosity > 1:
            outgoing = (
                Direction.LOCAL if direction is Direction.REMOTE else Direction.REMOTE
            )
            print(
                "{} Sent to {}".format(
                    get_direction_label(direction, self.__config.name), outgoing.name
                )
            )

    def __receive_from(self, connection: socket):
        buffer = b""
        if not connection:
            return buffer

        connection.settimeout(0.1)
        try:

            while True:
                chunk = connection.recv(4096)
                if not chunk:
                    break

                buffer += chunk
            if not len(buffer):
                self.stop()

        except socket.timeout:
            pass
        except socket.error as e:
            print(Fore.RED + e + Fore.RESET)
            pass
        return buffer

    def __request_handler(self, buffer: str):
        # perform any modifications bound for the remote host here
        return buffer

    def __response_handler(self, buffer: str):
        # perform any modifictions bound for the local host here
        return buffer



import setuptools

__version__ = None
with open("aproxy/version.py") as f:
    exec(f.read())

setuptools.setup(
    name="aproxy",
    version=__version__,
    packages=setuptools.find_packages(),
    entry_points={"console_scripts": ["aproxy=aproxy.__main__:main"]},
    author="Frode Hus",
    author_email="frode.hus@outlook.com",
    description="Simple tool that lets you run multiple proxies simultaneously",
    url="https://github.com/frodehus/aproxy",
    python_requires=">=3.6",
    install_requires=["hexdump", "colorama", "knack",],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
